from socket import gethostbyname
from time import sleep as wait

import socket

# from .secure_connection import SecureConnection
from .connection import Connection as SecureConnection
from .event import Event

WAIT_TIME = 0.2


class Client:
    def __init__(self, ip, port):
        self.IP = gethostbyname(ip)
        self.Port = port

        self.Connected = Event()
        self.Disconnected = Event()
        self.Signalled = Event()

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._connection = None

    def connect(self):
        self._socket.connect((self.IP, self.Port))
        self._connection = SecureConnection(self._socket)
        self._connection.listen()

        wait(WAIT_TIME)
        self.Connected.fire(self._connection)

        self._connection.Disconnected.connect(lambda: self.Disconnected.fire())
        
    def disconnect(self):
        self._connection.disconnect()



def setup_server(s):
    @s.Connected
    def handle_connection_on_server(conn):
        @conn.Signalled("/greet_server")
        def say_hello_on_server(signal):
            print("Hello, World!")

            conn.send(Signal("/respond_client", {
                "name": "John"
            }))


def setup_client(c):
    @c.Connected
    def handle_connection_on_client(conn):
        @conn.Signalled("/respond_client")
        def handle_response(signal):
            name = signal.data["name"]

            print(f"Greetings from {name}")

        conn.send(Signal("/greet_server"))


def main():
    s = Server(PORT)
    c = Client(IP, PORT)

    setup_server(s)
    setup_client(c)

    s.start()
    c.connect()

    wait(2)

    s.stop()


if __name__ == "__main__":
    from lib import Server, Client, Signal, get_host
    from time import sleep as wait

    IP = get_host()
    PORT = 7092

    main()



from threading import Thread, _active
from ctypes import pythonapi, py_object

class Parallel(Thread):
    def __init__(self, callback):
        super().__init__(daemon=False)

        self._callback = callback

    def run(self):
        try:
            self._callback()

        finally:
            return

    def get_id(self):
        if hasattr(self, "_thread_id"):
            return self._thread_id

        else:
            for id, thread in _active.items():
                if thread is self:
                    return id

    def cancel(self):
        t_id = self.get_id()

        pythonapi.PyThreadState_SetAsyncExc(t_id, py_object(SystemExit))



from threading import Event as Waiter
from types import FunctionType


class Event:
    def __init__(self):
        self._callbacks = []
        self._waiters = []
        self._waiter_return = None

    def __call__(self, fn):
        self.connect(fn)

    def fire(self, *args):
        for callback in self._callbacks:
            callback(*args)

        for w in self._waiters:
            self._waiter_return = args
            w.set()

        self._waiters = []

    def connect(self, fn):
        self._callbacks.append(fn)

    def wait(self):
        w = Waiter()

        self._waiters.append(w)

        w.wait()

        return self._waiter_return


class ConditionalEvent(Event):
    def __init__(self, checker, default=None):
        super().__init__()

        self._default = default
        self._callbacks = {}
        self._waiters = {}
        self._conditioner = checker

    def __call__(self, condition=None):
        if condition is None:
            condition = self._default

        return self.connect(condition)

    def fire(self, *args):
        target_condition = self._conditioner(*args)

        for condition, callback in self._callbacks.items():
            if condition in target_condition:
                callback(*args)

        for condition, waiter in self._waiters:
            if condition in target_condition:
                self._waiter_return = args
                waiter.set()
                self._waiter_return = None

    def connect(self, condition=None):
        if condition is None:
            condition = self._default

        def _handle_connect(fn: FunctionType):
            self._callbacks[condition] = fn

            return fn

        return _handle_connect

    def wait(self, condition):
        w = Waiter()

        if condition in self._waiters:
            self._waiters[condition].append(w)

        else:
            self._waiters[condition] = [w]

        w.wait()
        
        return self._waiter_return



from json import dumps, loads
from base64 import b64encode, b64decode


class Signal:
    ENCODING = "utf-8"

    def __init__(self, path="*", data=None):
        self.path = path
        self.data = data or {}

    @staticmethod
    def decode(encoded: bytes):
        decoded = loads(b64decode(encoded).decode(Signal.ENCODING))
        path = decoded["path"]
        data = decoded["data"]

        return Signal(path, data)

    def encoded(self):
        return bytes(self)

    def __bytes__(self):
        return b64encode(dumps({
            "path": self.path,
            "data": self.data
        }).encode(Signal.ENCODING))



import socket

from .event import Event, ConditionalEvent
from .signal import Signal
from .parallel import Parallel


class Connection:
    TERMINATOR = b"\n"
    CHUNK_SIZE = 1024

    def __init__(self, sock: socket.socket):
        self.ip = sock.getpeername()
        self._socket = sock

        self.Signalled = ConditionalEvent(lambda signal: [signal.path], default="*")
        self.Disconnected = Event()

        self.connected = False
        self._event_loop = Parallel(self._listen)

    def listen(self):
        self.connected = True

        self._event_loop.start()

    def disconnect(self):
        self.send(Signal("/__disconnect"))
        self._close()

    def _close(self):
        self.connected = False
        self.Disconnected.fire()

        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()
        self._event_loop.cancel()

    def _listen(self):
        while self.connected:
            stream = b""

            while Connection.TERMINATOR not in stream:
                stream += self._socket.recv(Connection.CHUNK_SIZE)

            packet = Signal.decode(stream.split(Connection.TERMINATOR)[0])

            if packet.path == "/__disconnect":
                self._close()

            else:
                self.Signalled.fire(packet)

    def send(self, data):
        encoded = bytes(data) + Connection.TERMINATOR

        size = len(encoded)
        current_chunks = 0
        while current_chunks < size:
            chunked_data = encoded[current_chunks:current_chunks + Connection.CHUNK_SIZE]
            self._socket.send(chunked_data)
            current_chunks += Connection.CHUNK_SIZE



from socket import gethostbyname, gethostname
import socket
import ssl

# from .secure_connection import SecureConnection
from typing import List

from .connection import Connection as SecureConnection, Connection
from .parallel import Parallel
from .event import Event


class Server:
    BACKLOG = 7

    def __init__(self, port):
        self.IP = gethostbyname(gethostname())
        self.Port = port

        self.Connected = Event()
        self.Disconnected = Event()
        self.Stopped = Event()
        self.Started = Event()

        # self._context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        # self._socket = self._context.wrap_socket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._connections: List[Connection] = []

        self._event_loop = Parallel(self._loop)

        self.running = True

    def send_to_all(self, signal):
        for conn in self._connections:
            conn.send(signal)

    def start(self):
        self._socket.bind((self.IP, self.Port))
        self._socket.listen()

        self._event_loop.start()

    def stop(self):
        self.running = False

        for conn in self._connections:
            if conn.connected is True:
                conn.disconnect()

        self._event_loop.cancel()

        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()

    def _loop(self):
        while self.running is True:
            try:
                sock, info = self._socket.accept()

                connection = SecureConnection(sock)
                connection.listen()
                self._connections.append(connection)

                connection.Disconnected.connect(lambda: self.Disconnected.fire(connection))

                self.Connected.fire(connection)

            except ConnectionResetError as e:
                continue



from setuptools import setup, find_packages

setup(
    name="signalio",
    version="1.0.3",
    description="A simple Python networking package",
    url="https://github.com/suchasaltylemon/Signal-IO",
    author="SuchASaltyLemon",
    author_email="suchasaltylemon@mailbox.org",
    packages=find_packages("src"),
    package_dir={"": "src"}
)



import socket
import ssl

from .connection import Connection


class SecureConnection(Connection):
    def __init__(self, sock: socket.socket):
        super().__init__(sock)

        self._context = ssl.SSLContext()
        self._socket = self._context.wrap_socket(sock)

    def send(self, data):
        pass
